import inspect
from copy import deepcopy
import numpy as np

from genepro.node import Node
from genepro import node_impl
from genepro.node_impl import Feature, Constant


def compute_linear_scaling(y,p):
  """
  Computes the optimal slope and intercept that realize the affine transformation that minimizes the mean-squared-error between the label and the prediction.
  See the paper: https://doi:10.1023/B:GENP.0000030195.77571.f9

  Parameters
  ----------
  y : np.array
    the label values
  p : np.array
    the respective predictions

  Returns
  -------
  float, float
    slope and intercept that represent 
  """
  slope = np.cov(y, p)[0,1] / (np.var(p) + 1e-12)
  intercept = np.mean(y) - slope*np.mean(p)
  return slope, intercept


def tree_from_prefix_repr(prefix_repr : str) -> Node:
  """
  Creates a tree from a string representation in prefix format (that is, pre-order tree traversal);
  the symbol in the string representation need to match those in the Node implementations (in `genepro.node_impl.py`)
  
  Parameters
  ----------
  prefix_repr : str
    the string representation of the tree as a list of nodes parsed with pre-order traversal (obtainable with `str(tree.get_subtree())`)

  Returns
  -------
  Node
    the tree that corresponds to the string representation
  """
  symb_list = prefix_repr.replace("[","").replace("]","").replace(", ",",").split(",")
  # generate the tree
  node_classes = [c[1] for c in inspect.getmembers(node_impl, inspect.isclass)]
  possible_nodes = list()
  for node_cls in node_classes:
    # handle Features and Constants separetely (also, avoid base class Node)
    if node_cls == Node or node_cls == Feature or node_cls == Constant:
      continue
    node_obj = node_cls()
    possible_nodes.append(node_obj)
  tree, _ = __tree_from_symb_list_recursive(symb_list, possible_nodes)
  return tree

def __tree_from_symb_list_recursive(symb_list : list, possible_nodes : list):
  """
  Helper recursive function for `tree_from_prefix_repr`

  Parameters
  ----------
  symb_list : list
    list of str that are symbols (as per the attribute `symb` of Node)

  possible_nodes : list
    list of all possible Node objects from `genepro.node_impl`

  Returns
  -------
    Node, list
      the last-generated node and the updated symb_list, required for the recursive construction of the tree
  """
  symb = symb_list[0]
  symb_list = symb_list[1:]
  # check if it is a feature
  if symb.startswith("x_"):
    id = int(symb[2:])
    n = Feature(id)
    return n, symb_list

  # check if it is a function
  for pn in possible_nodes:
    if symb == str(pn):
      n = deepcopy(pn)
      for _ in range(n.arity):
        c, symb_list = __tree_from_symb_list_recursive(symb_list, possible_nodes)
        n.insert_child(c)
      return n, symb_list

  # if reached this line, it must be a constant
  n = Constant(float(symb))
  return n, symb_list


import numpy as np
from numpy.random import random as randu
from numpy.random import normal as randn
from numpy.random import choice as randc
from numpy.random import shuffle
from copy import deepcopy

from genepro.node import Node
from genepro.node_impl import Constant


def generate_random_tree(internal_nodes : list, leaf_nodes : list, max_depth : int, curr_depth : int=0):
  """
  Recursive method to generate a random tree containing the given types of nodes and up to the given maximum depth

  Parameters
  ----------
  internal_nodes : list
    list of nodes with arity > 0 (i.e., sub-functions or operations)
  leaf_nodes : list
    list of nodes with arity==0 (also called terminals, e.g., features and constants)
  max_depth : int
    maximum depth of the tree (recall that the root node has depth 0)
  curr_depth : int
    the current depth of the tree under construction, it is set by default to 0 so that calls to `generate_random_tree` need not specify it

  Returns
  -------
  Node
    the root node of the generated tree
  """

  # heuristic to generate a semi-normal centered on relatively large trees
  prob_leaf = 0.01 + (curr_depth / max_depth)**3

  if curr_depth == max_depth or randu() < prob_leaf:
    n = deepcopy(randc(leaf_nodes))
  else:
    n = deepcopy(randc(internal_nodes))

  for _ in range(n.arity):
    c = generate_random_tree(internal_nodes, leaf_nodes, max_depth, curr_depth+1)
    n.insert_child(c)

  return n

def subtree_crossover(tree : Node, donor : Node, unif_depth : int=True) -> Node:
  """
  Performs subtree crossover and returns the resulting offspring

  Parameters
  ----------
  tree : Node
    the tree that participates and is modified by crossover
  donor : Node
    the second tree that participates in crossover, it provides candidate subtrees
  unif_depth : bool, optional
    whether uniform random depth sampling is used to pick the root of the subtrees to swap (default is True)

  Returns
  -------
  Node
    the tree after crossover (warning: replace the original tree with the returned one to avoid undefined behavior)
  """
  # pick a subtree to replace
  n = __sample_node(tree, unif_depth)
  m = deepcopy(__sample_node(donor, unif_depth))

  # remove ref to parent of m
  m.parent = None
  # swap
  p = n.parent
  if p:
    i = p.detach_child(n)
    p.insert_child(m,i)
  else:
    tree = m
  return tree

def node_level_crossover(tree : Node, donor : Node, same_depth : bool=False, prob_swap : float=0.1) -> Node:
  """
  Performs crossover at the level of single nodes

  Parameters
  ----------
  tree : Node
    the tree that participates and is modified by crossover
  donor : Node
    the second tree for crossover, which provides candidate nodes
  same_depth : bool, optional
    whether node-level swaps should occur only between nodes at the same depth level (default is False)
  prob_swap : float, optional
    the probability of swapping a node in tree with one in donor (default is 0.1)

  Returns
  -------
  Node
    the tree after crossover 
  """
  nodes = tree.get_subtree()
  donor_nodes = donor.get_subtree()

  donor_node_arity = dict()
  donor_node_arity_n_depth = dict()
  for n in donor_nodes:
    arity = n.arity
    if arity not in donor_node_arity:
      donor_node_arity[arity] = [n]
    else:
      donor_node_arity[arity].append(n)
    # also record depths if same_depth==True
    if same_depth:
      depth = n.get_depth()
      ar_n_dep = (arity, depth)
      if ar_n_dep not in donor_node_arity_n_depth:
        donor_node_arity_n_depth[ar_n_dep] = [n]
      else:
        donor_node_arity_n_depth[ar_n_dep].append(n)

  for n in nodes:
    if randu() < prob_swap:
      # find compatible donor
      arity = n.arity
      if same_depth:
        depth = n.get_depth()
        compatible_nodes = donor_node_arity_n_depth[(arity,depth)] if (arity,depth) in donor_node_arity_n_depth else None
      else:
        compatible_nodes = donor_node_arity[arity] if arity in donor_node_arity else None
      # if no compatible nodes, skip
      if compatible_nodes is None or len(compatible_nodes) == 0:
        continue
      # swap
      m = deepcopy(randc(compatible_nodes))
      m.parent = None
      m._children = list()
      p = n.parent
      if p:
        i = p.detach_child(n)
        p.insert_child(m,i)
      else:
        tree = m
      for c in n._children:
        m.insert_child(c)
  
  return tree


def subtree_mutation(tree : Node, internal_nodes : list, leaf_nodes : list, 
  unif_depth : bool=True, max_depth : int=4, prob_leaf : float=0.25) -> Node:
  """
  Performs subtree mutation and returns the resulting offspring

  Parameters
  ----------
  tree : Node
    the tree that undergoes mutation
  internal_nodes : list
    list of possible internal nodes to generate the mutated branch
  leaf_nodes : list
    list of possible leaf nodes to generate the mutated branch
  unif_depth : bool, optional
    whether uniform random depth sampling is used to pick the root of the subtree to mutate (default is True)
  max_depth : int, optional
    the maximal depth of the mutated branch (default is 4)
  prob_leaf : float, optional
    the probability of sampling a leaf when generating the mutated branch (default is 0.25)
  Returns
  -------
  Node
    the tree after mutation (warning: replace the original tree with the returned one to avoid undefined behavior)
  """
  # pick a subtree to replace
  n = __sample_node(tree, unif_depth)
  # generate a random branch
  branch = generate_random_tree(internal_nodes, leaf_nodes, max_depth, prob_leaf)
  # swap
  p = n.parent
  if p:
    i = p.detach_child(n)
    p.insert_child(branch,i)
  else:
    tree = branch
  return tree

def one_point_mutation(tree : Node, internal_nodes : list, leaf_nodes : list) -> Node:
  """
  Performs one-point mutation and returns the resulting offspring

  Parameters
  ----------
  tree : Node
    the tree that undergoes mutation
  internal_nodes : list
    list of possible internal nodes for replacing existing nodes
  leaf_nodes : list
    list of possible leaf nodes for replacing existing nodes
  Returns
  -------
  Node
    the tree after mutation (warning: replace the original tree with the returned one to avoid undefined behavior)
  """
  nodes = tree.get_subtree()
  prob_mutation = 1.0 / len(nodes)
  
  for n in tree.get_subtree():
    
    if randu() < prob_mutation:
      # get arity of current node, and check if leaf or not
      arity = n.arity
      is_leaf = arity == 0
      # pick a new node with the same arity
      if is_leaf:
        new_node = randc(leaf_nodes)
      else:
        possible_nodes = [m for m in internal_nodes if m.arity == arity]
        new_node = randc(possible_nodes)

      # attach children
      for c in n._children:
        new_node.insert_child(c)
      
      # link to parent
      p = n.parent
      if p:
        i = p.detach_child(n)
        p.insert_child(new_node,i)
      else:
        tree = new_node

    return tree
      
      
  
  # pick a subtree to replace
  n = __sample_node(tree, unif_depth)
  # generate a random branch
  branch = generate_random_tree(internal_nodes, leaf_nodes, max_depth, prob_leaf)
  # swap
  p = n.parent
  if p:
    i = p.detach_child(n)
    p.insert_child(branch,i)
  else:
    tree = branch
  return tree

def coeff_mutation(tree : Node, prob_coeff_mut : float= 0.25, temp : float=0.25) -> Node:
  """
  Applies random coefficient mutations to constant nodes 

  Parameters
  ----------
  tree : Node
    the tree to which coefficient mutations are applied
  prob_coeff_mut : float, optional
    the probability with which coefficients are mutated (default is 0.25)
  temp : float, optional
    "temperature" that indicates the strength of coefficient mutation, it is relative to the current value (i.e., v' = v + temp*abs(v)*N(0,1))

  Returns
  -------
  Node
    the tree after coefficient mutation (it is the same as the tree in input)
  """
  coeffs = [n for n in tree.get_subtree() if type(n) == Constant]
  for c in coeffs:
    # decide wheter it should be applied
    if randu() < prob_coeff_mut:
      v = c.get_value()
      # update the value by +- temp relative to current value
      new_v = v + temp*np.abs(v)*randn()
      c.set_value(new_v)
  
  return tree

def __sample_node(tree : Node, unif_depth : bool=True) -> Node:
  """
  Helper method that samples a random node from a tree

  Parameters
  ----------
  tree : Node
    the tree from which a random node should be sampled
  unif_depth : bool, optional
    whether the depth of the random node should be sampled uniformly at random first (default is True)

  Returns
  -------
  Node
    the randomly sampled node
  """
  nodes = tree.get_subtree()
  if unif_depth:
    nodes = __sample_uniform_depth_nodes(nodes)
  return randc(nodes)
  
def __sample_uniform_depth_nodes(nodes : list) -> list:
  """
  Helper method for `__sample_node` that returns candidate nodes that all have a depth which was sampled uniformly at random

  Parameters
  ----------
  nodes : list
    list of nodes from which to sample candidates that share a random depth (typically the result of `get_subtree()`)

  Returns
  -------
  list:
    list of nodes that share a depth that was sampled uniformly at random
  """
  depths = [n.get_depth() for n in nodes]
  possible_depths = np.unique(depths)
  d = randc(possible_depths)
  candidates = [n for i, n in enumerate(nodes) if depths[i] == d]
  return candidates
  



def generate_offspring(parent : Node, 
  crossovers : list, mutations : list, coeff_opts : list,
  donors : list, internal_nodes : list, leaf_nodes : list,
  constraints : dict={"max_tree_size": 100}) -> Node:
  """
  Generates an offspring from a given parent (possibly using a donor from the population for crossover).
  Variation operators are applied in a random order.
  The application of the variation operator is handled by `__undergo_variation_operator`

  Parameters
  ----------
  parent : Node
    the parent tree from which the offspring is generated by applying the variation operators
  crossovers : list
    list of dictionaries each specifying a type of crossover and respective hyper-parameters
  mutations : list
    list of dictionaries each specifying a type of mutation and respective hyper-parameters
  coeff_opts : list
    list of dictionaries each specifying a type of coefficient optimization and respective hyper-parameters
  donors : list
    list of Node, each representing a donor tree that can be used by crossover
  internal_nodes : list
    list of internal nodes to be used by mutation
  leaf_nodes : list
    list of internal nodes to be used by mutation
  constraints : dict, optional
    constraints the generated offspring must meet (default is {"max_size": 100})

  Returns
  -------
  Node
    the offspring after having applied the variation operators
  """
  # set the offspring to a copy (to be modified) of the parent
  offspring = deepcopy(parent)
  # create a backup for constraint violation
  backup = deepcopy(offspring)

  # apply variation operators in a random order
  all_var_ops = crossovers + mutations + coeff_opts
  random_order = np.arange(len(all_var_ops))
  shuffle(random_order)
  for i in random_order:
    var_op = all_var_ops[i]
    offspring = __undergo_variation_operator(var_op, offspring, 
      crossovers, mutations, coeff_opts,
      randc(donors), internal_nodes, leaf_nodes)
    # check offspring meets constraints, else revert to backup
    if not __check_tree_meets_all_constraints(offspring, constraints):
      # revert to backup
      offspring = deepcopy(backup)
    else:
      # update backup
      backup = deepcopy(offspring)

  return offspring


def __undergo_variation_operator(var_op : dict, offspring : Node,
  crossovers, mutations, coeff_opts,
  donor, internal_nodes, leaf_nodes) -> Node:
  # decide whether to actually do something
  if var_op["rate"] < randu():
    # nope
    return offspring

  # prepare the function to call
  var_op_fun = var_op["fun"]
  # next, we need to provide the right arguments based on the type of ops
  if var_op in crossovers:
    # we need a donor
    offspring = var_op_fun(offspring, donor, **var_op["kwargs"])
  elif var_op in mutations:
    # we need to provide node types 
    offspring = var_op_fun(offspring, internal_nodes, leaf_nodes, **var_op["kwargs"])
  elif var_op in coeff_opts:
    offspring = var_op_fun(offspring, **var_op["kwargs"])

  return offspring


def __check_tree_meets_all_constraints(tree : Node, constraints : dict=dict()) -> bool:
  """
  """
  meets = True
  for constraint_name in constraints.keys():
    if constraint_name == "max_tree_size":
      if len(tree.get_subtree()) > constraints["max_tree_size"]:
        meets = False
        break
    else:
      raise ValueError("Unrecognized constraint name: {}".format(constraint_name))
  return meets



from __future__ import annotations
import numpy as np

class Node:
  """
  Class to represent a node in a tree.
  When used to represent the root, it also represents the tree.
  Implementations (in `genepro.node_impl`) inherit from this class and extend it to represent operations, features, or constants.

  Attributes
  ----------
  symb : str
    the symbol that represents the function of this node (e.g., "+" for Plus)
  fitness : float
    the fitness of the tree, only meaningful for the root note
  parent : Node
    the parent node of this node, it is None for the root note
  arity : int
    number of input arguments this node accepts (e.g., 2 for Plus, 1 for Log)
  _children : list
    list of nodes that are children whose parent is this node (warning: do not write on this field directly, use `insert_child` or `detach_child`)
  """
  def __init__(self):
    self.symb = None
    self.fitness = -np.inf
    self.parent = None
    self.arity = 0
    self._children = list()
  
  def __repr__(self) -> str:
    """
    Returns the string representation of this node

    Returns
    -------
    str
      the attribute symb
    """
    return self.symb

  def __len__(self) -> int:
    """
    Returns the length of the subtree rooted at this node
  
    Returns
    -------
      length of the subtree rooted at this node
    """
    return len(self.get_subtree())

  def __call__(self, X : np.ndarray) -> np.ndarray:
    """
    Shorthand for get_output

    Parameters
    ----------
    X : np.ndarray
      the input to be processed as an numpy ndarray with dimensions (num. observations, num. features)
    
    Returns
    -------
    np.ndarray
      the output obtained by processing the input
    """
    return self.get_output(X)

  def get_subtree(self) -> list:
    """
    Returns the subtree rooted at this node, as a list of nodes visited in prefix order

    Returns
    -------
    list
      the subtree (including this node) as a list of descendant nodes in prefix order
    """
    subtree = list()
    self.__get_subtree_recursive(subtree)
    return subtree

  def get_readable_repr(self) -> str:
    """
    Builds a human-readable representation of the subtree

    Returns
    -------
    str
      Human-readable representation of the subtree
    """
    repr = [""] # trick to pass string by reference
    self.__get_readable_repr_recursive(repr)
    return repr[0]

  def insert_child(self, c: Node, i: int=None):
    """
    Inserts the given node among the children of this node

    Parameters
    ----------
    c : Node
      the node to insert as a child of this node
    i : int, optional
      the position at which to insert c (default is None, in which case c is appended)
    """
    if i is None:
      self._children.append(c)
    else:
      self._children.insert(i,c)
    c.parent=self

  def detach_child(self, c : Node) -> int:
    """
    Removes the given node from the children of this node and returns its position relative to the other children

    Parameters
    ----------
    c : Node
      the node to detach
    
    Returns
    -------
    int
      index specifying the position of c in the attribute _children
    """
    assert(c in self._children)
    for i, oc in enumerate(self._children):
      if c == oc:
        self._children.pop(i)
        c.parent = None
        break
    return i

  def get_output(self, X : np.ndarray) -> np.ndarray: 
    """
    Returns the output of this node after processing the given input

    Parameters
    ----------
    X : np.ndarray
      the input to be processed as an numpy ndarray with dimensions (num. observations, num. features)
    
    Returns
    -------
    np.ndarray
      the output obtained by processing the input
    """
    raise NotImplementedError()

  def get_depth(self) -> int:
    """
    Returns the depth of this node (the root node has depth 0)

    Returns
    -------
    int
      the depth of this node
    """
    n = self
    d = 0
    while n.parent:
      d += 1
      n = n.parent
    return d

  def get_height(self) -> int:
    """
    Computes and returns the height of this node

    Returns
    -------
    int
      the height of this node
    """
    curr_depth = self.get_depth()
    leaves = [x for x in self.get_subtree() if x.arity == 0]
    max_h = 0
    for l in leaves:
      d = l.get_depth()
      if d > max_h:
        max_h = d
    return max_h - curr_depth

  def _get_child_outputs(self, X : np.ndarray) -> list:
    """
    Returns the output of the children for the given input as a list

    Parameters
    ----------
    X : np.ndarray
      the input to be processed as an numpy ndarray with dimensions (num. observations, num. features)
    
    Returns
    -------
    list
      list containing the output of the children, each as a numpy.ndarray
    """
    c_outs = list()
    for i in range(self.arity):
      c_o = self._children[i].get_output(X)
      c_outs.append(c_o)
    return c_outs

  def _get_args_repr(self, args : list) -> str:
    """
    Returns a string representing the way this node processes its input arguments (i.e., its children)

    Parameters
    ----------
    args : list
      list of strings, each representing the way a child node processes its input arguments
    
    Returns
    -------
    str
      representation of the way this node processes its input arguments
    """
    raise NotImplementedError()

  def _get_typical_repr(self, args : list, name : str) -> str:
    """
    Helper method for `_get_args_repr` implementing typical representations

    Parameters
    ----------
    args : list
      list of strings, each representing the way a child node processes its input arguments
    name : str, options are "between", "before", and "after"
      name of the typical representation, needs to match with the length of args (e.g., "between" for Plus results in "args[0] + args[1]")
    
    Returns
    -------
    str
      typical representation of the way this node processes its input arguments
    """
    if name == 'between':
      if len(args) != 2:
        raise ValueError("Invalid representation 'between' for len(args)!=2")
      return '('+args[0]+self.symb+args[1]+')'
    elif name == 'before':
      repr = self.symb+'('
      for arg in args:
        repr += arg + ','
      repr = repr[:-1] + ')'
      return repr
    elif name == 'after':
      repr = '('
      for arg in args:
        repr += arg + ','
      repr = repr[:-1] + ')'
      repr += self.symb
      return repr
    else:
      raise ValueError("Unrecognized option {}".format(name))
  
  def __get_subtree_recursive(self, subtree : list):
    """
    Helper method for `get_subtree` that uses recursion to visit the descendant nodes and populate the given list

    Parameters
    ----------
    subtree : list
      list that is populated by including this node and then calling this method again on the children of that node
    """
    subtree.append(self)
    for c in self._children:
      c.__get_subtree_recursive(subtree)

  def __get_readable_repr_recursive(self, repr : list):
    """
    Helper method for get_subtree_repr that uses recursion to visit the descendant nodes and populate the given list

    Parameters
    ----------
    repr : list
      list that is used as container to fill a string with the result of `_get_args_repr` of this node and its descendants
    """
    args = list()
    for i in range(self.arity):
      self._children[i].__get_readable_repr_recursive(repr)
      args.append(repr[0])
    repr[0] = self._get_args_repr(args)




from typing import Callable

import numpy as np
from numpy.random import random as randu
from numpy.random import randint as randi
from numpy.random import choice as randc
from numpy.random import shuffle
import time, inspect
from copy import deepcopy
from joblib.parallel import Parallel, delayed

from genepro.node import Node
from genepro.variation import *
from genepro.selection import tournament_selection

class Evolution:
  """
  Class concerning the overall evolution process.

  Parameters
  ----------
  fitness_function : function
    the function used to evaluate the quality of evolving trees, should take a Node and return a float; higher fitness is better

  internal_nodes : list
    list of Node objects to be used as internal nodes for the trees (e.g., [Plus(), Minus(), ...])

  leaf_nodes : list
    list of Node objects to be used as leaf nodes for the trees (e.g., [Feature(0), Feature(1), Constant(), ...])

  pop_size : int, optional
    the population size (default is 256)

  init_max_depth : int, optional
    the maximal depth trees can have at initialization (default is 4)

  max_tree_size : int, optional
    the maximal number of nodes trees can have during the entire evolution (default is 64)

  crossovers : list, optional
    list of dictionaries that contain: "fun": crossover functions to be called, "rate": rate of applying crossover, "kwargs" (optional): kwargs for the chosen crossover function (default is [{"fun":subtree_crossover, "rate": 0.75}])

  mutations : list, optional
    similar to `crossovers`, but for mutation (default is [{"fun":subtree_mutation, "rate": 0.75}])

  coeff_opts : list, optional
    similar to `crossovers`, but for coefficient optimization (default is [{"fun":coeff_mutation, "rate": 1.0}])
  
  selection : dict, optional
    dictionary that contains: "fun": function to be used to select promising parents, "kwargs": kwargs for the chosen selection function (default is {"fun":tournament_selection,"kwargs":{"tournament_size":4}})

  max_evals : int, optional
    termination criterion based on a maximum number of fitness function evaluations being reached (default is None)

  max_gens : int, optional
    termination criterion based on a maximum number of generations being reached (default is 100)

  max_time: int, optional
    termination criterion based on a maximum runtime being reached (default is None)

  n_jobs : int, optional
    number of jobs to use for parallelism (default is 4)

  verbose : bool, optional
    whether to log information during the evolution (default is False)

  Attributes
  ----------
  All of the parameters, plus the following:

  population : list
    list of Node objects that are the root of the trees being evolved

  num_gens : int
    number of generations

  num_evals : int
    number of evaluations

  start_time : time
    start time

  elapsed_time : time
    elapsed time

  best_of_gens : list
    list containing the best-found tree in each generation; note that the entry at index 0 is the best at initialization
  """
  def __init__(self,
    # required settings
    fitness_function : Callable[[Node], float],
    internal_nodes : list,
    leaf_nodes : list,
    # optional evolution settings
    pop_size : int=256,
    init_max_depth : int=4,
    max_tree_size : int=64,
    crossovers : list=[{"fun":subtree_crossover, "rate": 0.5}],
    mutations : list= [
      {"fun":subtree_mutation, "rate": 0.5},
      {"fun":one_point_mutation, "rate": 0.5},
    ],
    coeff_opts : list = [{"fun":coeff_mutation, "rate": 0.5}],
    selection : dict={"fun":tournament_selection,"kwargs":{"tournament_size":8}},
    # termination criteria
    max_evals : int=None,
    max_gens : int=100,
    max_time : int=None,
    # other
    n_jobs : int=4,
    verbose : bool=False,
    ):

    # set parameters as attributes
    _, _, _, values = inspect.getargvalues(inspect.currentframe())
    values.pop('self')
    for arg, val in values.items():
      setattr(self, arg, val)

    # fill-in empty kwargs if absent in crossovers, mutations, coeff_opts
    for variation_list in [crossovers, mutations, coeff_opts]:
      for i in range(len(variation_list)):
        if "kwargs" not in variation_list[i]:
          variation_list[i]["kwargs"] = dict()
    # same for selection
    if "kwargs" not in selection:
      selection["kwargs"] = dict()

    # initialize some state variables
    self.population = list()
    self.num_gens = 0
    self.num_evals = 0
    self.start_time, self.elapsed_time = 0, 0
    self.best_of_gens = list()


  def _must_terminate(self) -> bool:
    """
    Determines whether a termination criterion has been reached

    Returns
    -------
    bool
      True if a termination criterion is met, else False
    """
    self.elapsed_time = time.time() - self.start_time
    if self.max_time and self.elapsed_time >= self.max_time:
      return True
    elif self.max_evals and self.num_evals >= self.max_evals:
      return True
    elif self.max_gens and self.num_gens >= self.max_gens:
      return True
    return False

  def _initialize_population(self):
    """
    Generates a random initial population and evaluates it
    """
    # initialize the population
    self.population = Parallel(n_jobs=self.n_jobs)(
        delayed(generate_random_tree)(
          self.internal_nodes, self.leaf_nodes, max_depth=self.init_max_depth )
        for _ in range(self.pop_size))

    # evaluate the trees and store their fitness
    fitnesses = Parallel(n_jobs=self.n_jobs)(delayed(self.fitness_function)(t) for t in self.population)
    for i in range(self.pop_size):
      self.population[i].fitness = fitnesses[i]
    # store eval cost
    self.num_evals += self.pop_size
    # store best at initialization
    best = self.population[np.argmax([t.fitness for t in self.population])]
    self.best_of_gens.append(deepcopy(best))

  def _perform_generation(self):
    """
    Performs one generation, which consists of parent selection, offspring generation, and fitness evaluation
    """
    # select promising parents
    sel_fun = self.selection["fun"]
    parents = sel_fun(self.population, self.pop_size, **self.selection["kwargs"])
    # generate offspring
    offspring_population = Parallel(n_jobs=self.n_jobs)(delayed(generate_offspring)
      (t, self.crossovers, self.mutations, self.coeff_opts, 
      parents, self.internal_nodes, self.leaf_nodes,
      constraints={"max_tree_size": self.max_tree_size}) 
      for t in parents)

    # evaluate each offspring and store its fitness 
    fitnesses = Parallel(n_jobs=self.n_jobs)(delayed(self.fitness_function)(t) for t in offspring_population)
    for i in range(self.pop_size):
      offspring_population[i].fitness = fitnesses[i]
    # store cost
    self.num_evals += self.pop_size
    # update the population for the next iteration
    self.population = offspring_population
    # update info
    self.num_gens += 1
    best = self.population[np.argmax([t.fitness for t in self.population])]
    self.best_of_gens.append(deepcopy(best))

  def evolve(self):
    """
    Runs the evolution until a termination criterion is met;
    first, a random population is initialized, second the generational loop is started:
    every generation, promising parents are selected, offspring are generated from those parents, 
    and the offspring population is used to form the population for the next generation
    """
    # set the start time
    self.start_time = time.time()

    self._initialize_population()

    # generational loop
    while not self._must_terminate():
      # perform one generation
      self._perform_generation()
      # log info
      if self.verbose:
        print("gen: {},\tbest of gen fitness: {:.3f},\tbest of gen size: {}".format(
            self.num_gens, self.best_of_gens[-1].fitness, len(self.best_of_gens[-1])
            ))



import numpy as np
from genepro.node import Node


class Plus(Node):
  def __init__(self):
    super(Plus,self).__init__()
    self.arity = 2
    self.symb = '+'

  def _get_args_repr(self, args):
    return self._get_typical_repr(args,'between')

  def get_output(self, X):
    c_outs = self._get_child_outputs(X)
    return c_outs[0] + c_outs[1]


class Minus(Node):
  def __init__(self):
    super(Minus,self).__init__()
    self.arity = 2
    self.symb = '-'

  def _get_args_repr(self, args):
    return self._get_typical_repr(args,'between')

  def get_output(self, X):
    c_outs = self._get_child_outputs(X)
    return c_outs[0] - c_outs[1]


class Times(Node):
  def __init__(self):
    super(Times,self).__init__()
    self.arity = 2
    self.symb = '*'

  def _get_args_repr(self, args):
    return self._get_typical_repr(args,'between')

  def get_output(self, X):
    c_outs = self._get_child_outputs(X)
    return np.multiply(c_outs[0], c_outs[1])


class Div(Node):
  def __init__(self):
    super(Div,self).__init__()
    self.arity = 2
    self.symb = '/'

  def _get_args_repr(self, args):
    return self._get_typical_repr(args,'between')

  def get_output(self, X):
    c_outs = self._get_child_outputs(X)
    # implements a protection to avoid dividing by 0
    sign_b = np.sign(c_outs[1])
    sign_b = np.where(sign_b == 0, 1, sign_b) 
    protected_div = sign_b * c_outs[0] / (1e-9 + np.abs(c_outs[1]))
    return protected_div


class Square(Node):
  def __init__(self):
    super(Square,self).__init__()
    self.arity = 1
    self.symb = '**2'

  def _get_args_repr(self, args):
    return self._get_typical_repr(args,'after')

  def get_output(self, X):
    c_outs = self._get_child_outputs(X)
    return np.square(c_outs[0])


class Cube(Node):
  def __init__(self):
    super(Cube,self).__init__()
    self.arity = 1
    self.symb = '**3'

  def _get_args_repr(self, args):
    return self._get_typical_repr(args,'after')

  def get_output(self, X):
    c_outs = self._get_child_outputs(X)
    return np.multiply(np.square(c_outs[0]), c_outs[0])


class Sqrt(Node):
  def __init__(self):
    super(Sqrt,self).__init__()
    self.arity = 1
    self.symb = 'sqrt'

  def _get_args_repr(self, args):
    # let's report also protection
    return "sqrt(abs("+args[0]+"))"

  def get_output(self, X):
    c_outs = self._get_child_outputs(X)
    # implements a protection to avoid arg <= 0
    return np.sqrt(np.abs(c_outs[0]))


class Log(Node):
  def __init__(self):
    super(Log,self).__init__()
    self.arity = 1
    self.symb = 'log'

  def _get_args_repr(self, args):
    # let's report also protection (to some level of detail)
    return "log(abs("+args[0]+"))"

  def get_output(self, X):
    c_outs = self._get_child_outputs(X)
    # implements a protection to avoid arg <= 0
    protected_log = np.log(np.abs(c_outs[0]) + 1e-9)
    return protected_log


class Exp(Node):
  def __init__(self):
    super(Exp,self).__init__()
    self.arity = 1
    self.symb = "exp"

  def _get_args_repr(self, args):
    return self._get_typical_repr(args,'before')

  def get_output(self, X):
    c_outs = self._get_child_outputs(X)
    return np.exp(c_outs[0])


class Sin(Node):
  def __init__(self):
    super(Sin,self).__init__()
    self.arity = 1
    self.symb = "sin"

  def _get_args_repr(self, args):
    return self._get_typical_repr(args,'before')

  def get_output(self, X):
    c_outs = self._get_child_outputs(X)
    return np.sin(c_outs[0])


class Cos(Node):
  def __init__(self):
    super(Cos,self).__init__()
    self.arity = 1
    self.symb = "cos"

  def _get_args_repr(self, args):
    return self._get_typical_repr(args,'before')

  def get_output(self, X):
    c_outs = self._get_child_outputs(X)
    return np.cos(c_outs[0])


class Max(Node):
  def __init__(self):
    super(Max,self).__init__()
    self.arity = 2
    self.symb = "max"

  def _get_args_repr(self, args):
    return self._get_typical_repr(args,"before")

  def get_output(self, X):
    c_outs = self._get_child_outputs(X)
    return np.where(c_outs[0]>c_outs[1], c_outs[0], c_outs[1])


class Min(Node):
  def __init__(self):
    super(Min,self).__init__()
    self.arity = 2
    self.symb = "min"

  def _get_args_repr(self, args):
    return self._get_typical_repr(args,"before")

  def get_output(self, X):
    c_outs = self._get_child_outputs(X)
    return np.where(c_outs[0]<c_outs[1], c_outs[0], c_outs[1])


class IfThenElse(Node):
  def __init__(self):
    super(IfThenElse,self).__init__()
    self.arity = 3
    self.symb = "if-then-else"

  def _get_args_repr(self, args):
    return "if("+args[0]+" >= 0)then("+args[1]+")else("+args[2]+")"

  def get_output(self, X):
    c_outs = self._get_child_outputs(X)
    return np.where(c_outs[0]>=0, c_outs[1], c_outs[2])


class Feature(Node):
  def __init__(self,id):
    super(Feature,self).__init__()
    self.arity = 0
    self.id = id
    self.symb = 'x_'+str(id)

  def _get_args_repr(self, args):
    return self.symb

  def get_output(self, X):
    return X[:,self.id]

class Constant(Node):
  def __init__(self, value : float=None):
    super(Constant,self).__init__()
    self.arity = 0
    self.__value = value
    self.symb = str(value) if value is not None else "const?"

  def get_value(self):
    if not self.__value:
      # sample uniformly between -5 and +5
      self.__value = np.random.uniform()*10 - 5 
      self.symb = str(self.__value)
    return self.__value 

  def set_value(self, value : float):
    self.__value = value   
    self.symb = str(value)

  def _get_args_repr(self, args):
    # make sure it is initialized
    self.get_value()
    return self.symb

  def get_output(self, X : np.ndarray) -> np.ndarray:
    # make sure it is initialized
    v = self.get_value()
    return np.repeat(v, len(X))


import numpy as np
from sklearn.base import BaseEstimator
from sklearn.metrics import balanced_accuracy_score, mean_squared_error
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from sklearn.utils.multiclass import unique_labels

from genepro.evo import Evolution
from genepro.node_impl import *
from genepro.util import compute_linear_scaling

class GeneProEstimator(BaseEstimator):
  def __init__(self,
    score=None,
    evo_kwargs=dict(),
    ):

    self.score = score

    # set up default internal nodes if not provided
    if "internal_nodes" not in evo_kwargs:
      evo_kwargs["internal_nodes"] = [Plus(), Minus(), Times(), Div(), Square()]
    # default leaf nodes can only be set at fitting time (when we know X)
    # hence, initially set to none here, will be set when calling .fit(X,y)
    if "leaf_nodes" not in evo_kwargs:
      evo_kwargs["leaf_nodes"] = None

    self.evo = Evolution(
      fitness_function=self.score,
      **evo_kwargs
    )

  def fit(self, X, y):
    # check that X and y have correct shape
    X, y = check_X_y(X, y)
    
    self.X_ = X
    self.y_ = y

    # default generation of leaf nodes
    if not self.evo.leaf_nodes:
      self.evo.leaf_nodes = [Feature(i) for i in range(X.shape[1])] + [Constant()]

    pass

  def predict(self, X):
    # check is fit had been called
    check_is_fitted(self)

    # input validation
    X = check_array(X)

    pass


class GeneProRegressor(GeneProEstimator):
  def __init__(self,
    score=None,
    use_linear_scaling=True,
    evo_kwargs=dict(),
    ):
    super(GeneProRegressor,self).__init__(score, evo_kwargs)

    # set score to default if not provided
    if score is None:
      def neg_mse(y, p):
        return -mean_squared_error(y, p)
      self.score = neg_mse

    self.use_linear_scaling = use_linear_scaling

    # create a fitness function
    def fitness_function(tree):
      pred = tree(self.X_)
      if self.use_linear_scaling:
        slope, intercept = compute_linear_scaling(self.y_, pred)
        pred = intercept + slope*pred
      return self.score(self.y_, pred)

    self.evo.fitness_function = fitness_function
  
  def fit(self, X, y):
    super(GeneProRegressor,self).fit(X,y)
    self.evo.evolve()


  def predict(self, X, best_ever=False):
    super(GeneProRegressor,self).predict(X)
    if best_ever:
      best = self.evo.best_of_gens[np.argmax([t.fitness for t in self.evo.best_of_gens])]
    else:
      best = self.evo.best_of_gens[-1]
    pred = best(X)
    if self.use_linear_scaling:
      # compute linear scaling coefficients w.r.t. training set
      pred_ = best(self.X_)
      slope = np.cov(self.y_, pred_)[0,1] / (np.var(pred_) + 1e-12)
      intercept = np.mean(self.y_) - slope*np.mean(pred_)
      # update prediction with linear scaling coefficients
      pred = intercept + slope*pred
    return pred


class GeneProClassifier(GeneProEstimator):
  def __init__(self,
    score=None,
    evo_kwargs=dict(),
    ):
    super(GeneProClassifier,self).__init__(score, evo_kwargs)

    # set score to default if not provided
    if score is None:
      self.score = balanced_accuracy_score

    # create a fitness function
    def fitness_function(tree):
      out = tree(self.X_)
      pred = np.where(out < 0, -1, 1)
      return self.score(self.y_, pred)

    self.evo.fitness_function = fitness_function
  
  def fit(self, X, y):
    super(GeneProClassifier,self).fit(X,y)

    # store classes
    self.classes_ = unique_labels(self.y_)
    if len(self.classes_) > 2:
      raise ValueError("Only binary classification is supported (repeat a one vs. all approach for multi-class)")

    # convert y_ into -1 and +1
    self.y_ = np.where(self.y_ == self.classes_[0], -1, +1)

    self.evo.evolve()

  def predict(self, X, best_ever=False):
    super(GeneProClassifier,self).predict(X)
    if best_ever:
      best = self.evo.best_of_gens[np.argmax([t.fitness for t in self.evo.best_of_gens])]
    else:
      best = self.evo.best_of_gens[-1]
    out = best(X)
    pred = np.where(out < 0, -1, 1)
    pred = np.where(pred == -1, self.classes_[0], self.classes_[1])

    return pred


import numpy as np
from copy import deepcopy

def tournament_selection(contestants : list, num_to_select : int, tournament_size : int=4) -> list:
  """
  Performs tournament selection on the contestants until the given number of selected contestants is reached;
  note that `len(contestants)` needs to be a multiple of `tournament_size` and similarly for `num_to_select`

  Parameters
  ----------
  contestants : list
    list of Node containing trees that undergo the selection
  num_survivors : int
    how many should be selected
  tournament_size : int, optional
    the size (window) of tournament selection (default is 4)

  Returns
  -------
  list
    list containing (copies of) the trees that were selected
  """
  selected = list()

  n = len(contestants)
  num_selected_per_parse = n // tournament_size
  num_parses = num_to_select // num_selected_per_parse

  # assert quantities are compatible
  assert n / tournament_size == num_selected_per_parse, "Number of contestants {} is not a multiple of tournament size {}".format(n,tournament_size)
  assert num_to_select / num_selected_per_parse == num_parses
 
  for _ in range(num_parses):
    # shuffle
    np.random.shuffle(contestants)
    fitnesses = np.array([t.fitness for t in contestants])

    winning_indices = np.argmax(fitnesses.reshape((-1, tournament_size)), axis=1)
    winning_indices += np.arange(0, n, tournament_size)

    selected += [deepcopy(contestants[i]) for i in winning_indices]

  return selected
 